/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef OPENTHREADSIMULATOR_H
#define OPENTHREADSIMULATOR_H

#include <OpenThreads/Thread>
#include <OpenThreads/Mutex>
#include <OpenThreads/Condition>

#include <simulator.h>

struct TimerThread;
struct StartThread;

/**
    An implementation of the simulator facility using the OpenThread library for maximum portability

    @author Nicolas Brodu
*/
class OpenThreadSimulator : public Simulator
{
public:
    OpenThreadSimulator();

    ~OpenThreadSimulator();

    virtual void start(bool threaded = true);

    virtual double getSystemTime();

    virtual void eventMutexLock();
    
    virtual void eventMutexUnlock();

    virtual bool stepControlMutexTryLock();

    virtual void stepControlMutexLock();
    
    virtual void stepControlMutexUnlock();
    
    virtual void waitSemaphorePost();
    
    virtual void waitSemaphoreFlush();
    
    virtual void waitSemaphoreWait();

    virtual void startWaitTimer(double delay);
    
    virtual void cancelWaitTimer();

protected:
    OpenThreads::Mutex eventMutex;
    OpenThreads::Mutex stepControlMutex;

    // OpenThread does not provide a semaphore facility
    // Implement what we need with other elements
    int waitSemaphoreCounter;
    OpenThreads::Mutex waitSemaphoreMutex;
    OpenThreads::Condition waitSemaphoreCondition;

    // The timer thread. See cpp file
    friend struct TimerThread;
    friend struct StartThread;
    TimerThread* timerThread;
    StartThread* startThread;
};

#endif
