/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef POSIX_SIMULATOR_H
#define POSIX_SIMULATOR_H

#include <simulator.h>

#include <pthread.h>
#include <semaphore.h>

/**

Implements the thread, semaphore, timer and mutex functions
used in the base Simulator class using POSIX functions.

@author Nicolas Brodu
*/
class POSIX_Simulator : public Simulator
{
public:
    POSIX_Simulator();

    ~POSIX_Simulator();
    
    virtual void start(bool threaded = true);
    
    virtual double getSystemTime();
    
protected:

    pthread_mutex_t event_mutex, control_mutex;
    sem_t wait_semaphore;
    
    pthread_mutex_t timer_mutex;
    pthread_cond_t timer_condition;
    
    // Implement our own semaphore because the cost of delete/create in flushing is too much with pthread
    pthread_mutex_t waitSemaphoreMutex;
    pthread_cond_t waitSemaphoreCondition;
    int waitSemaphoreCounter;

    virtual void eventMutexLock();
    
    virtual void eventMutexUnlock();

    virtual bool stepControlMutexTryLock();

    virtual void stepControlMutexLock();
    
    virtual void stepControlMutexUnlock();
    
    virtual void waitSemaphorePost();
    
    virtual void waitSemaphoreFlush();
    
    virtual void waitSemaphoreWait();

    virtual void startWaitTimer(double delay);
    
    virtual void cancelWaitTimer();

    double waitDelay;
    friend void* timer_routine(void* arg);
    friend void* run_routine(void* arg);
};

#endif
