/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "DrawableSporeCloud.h"
#include "Environment.h"
using namespace Environment;
/*
#include <GL/gl.h>
#include <GL/glext.h>

#if defined(_WIN32) || defined(WIN32)
// ???
#else
#include <GL/glx.h>
#endif
*/

#include <osg/GL>
#include <osg/GLExtensions>
#include <osg/VertexProgram>
#include <osg/StateSet>
#include <osg/Texture2D>
#include <osg/BlendFunc>
#include <osg/TexEnv>
#include <osg/Material>

#include <osgDB/ReadFile>

#include <iostream>
using namespace std;

/*struct FactoryHelper {
    FactoryHelper(int size) {
        array = new DrawableSporeCloud::ColoredVolume[size];
        counter = 0;
    }
    DrawableSporeCloud::ColoredVolume* array;
    int counter;
};
*/

// Factory for parent to create objects of the correct type
SporeCloud::SporeVolume * DrawableSporeCloud::coloredVolumeFactory( void * user )
{
/*    FactoryHelper* helper = static_cast<FactoryHelper*>(user);
    return &helper->array[helper->counter++];
*/
    return new ColoredVolume;
}

DrawableSporeCloud::DrawableSporeCloud( Terrain * terrain, int divx, int divy, int divz, double baseDiffusionCoef, double lowestConcentration )
: osg::Drawable(), 
SporeCloud(-terrain->getHalfWorldSize(), -terrain->getHalfWorldSize(), 0.0,
             terrain->getWorldSize(), terrain->getWorldSize(), terrain->getMaxHeight(),
             divx, divy, divz, baseDiffusionCoef, lowestConcentration, &coloredVolumeFactory)//, new FactoryHelper(divx*divy*divz))
{
    // Start with no visible quad, will be updated at next step
    visibleList = 0;
    viewDisabled = false;

    setSupportsDisplayList(false);
    stepx = sizex / (double)divx;
    stepy = sizey / (double)divy;
    stepz = sizez / (double)divz;
    
    // take max of x-y-z steps, and sqrt(3) for the cube diagonal
    // than add 1/4 each so 1/2 overlap.
    billboardWidth = stepx;
    if (billboardWidth < stepy) billboardWidth = stepy;
    if (billboardWidth < stepz) billboardWidth = stepz;
    billboardWidth *= 1.7320508 * 1.25;
    billboardHeight = billboardWidth;
//    billboardWidth = 5.0; // Debug center position
//    billboardHeight = 5.0;
    halfBillboardWidth = billboardWidth * 0.5;
    halfBillboardHeight = halfBillboardWidth;

    // Initialize color & positions
    SporeVolume** it = volume;
    float xpos, ypos, zpos;
    float xposinit, yposinit;
    xposinit = originx + 0.5 * stepx;
    yposinit = originy + 0.5 * stepy;
    zpos = originz + 0.5 * stepz;
    
    //cout << "terrain->dsize=" << terrain->getDataArraySize() << endl;
    double dxt = sizex / (double)terrain->getDataArraySize();
    double dyt = sizey / (double)terrain->getDataArraySize();
    
    SporeVolume* prev = 0;
    
    for (int z=0; z<divz; ++z) {
        ypos=yposinit;
        for (int y=0; y<divy; ++y) {
            xpos=xposinit;
            for (int x=0; x<divx; ++x) {

                double dxmax = originx + (x+1.0)*stepx;
                double dymax = originy + (y+1.0)*stepy;
                double ox = originx + x*stepx;
                double oy = originy + y*stepy;
                double minz = terrain->getMaxHeight();
                double maxz = 0.0;
                double height;
                for (double dx = ox; dx <= dxmax; dx += dxt) {
                    for (double dy = oy; dy <= dymax; dy += dyt) {
                        height = terrain->getFloorZ(dx, dy);
                        if (height < minz) minz = height;
                        if (height > maxz) maxz = height;
                    }
                }
                // process remaining corners
                height = terrain->getFloorZ(ox, dymax);
                if (height < minz) minz = height;
                if (height > maxz) maxz = height;
                height = terrain->getFloorZ(dxmax, oy);
                if (height < minz) minz = height;
                if (height > maxz) maxz = height;
                height = terrain->getFloorZ(dxmax, dymax);
                if (height < minz) minz = height;
                if (height > maxz) maxz = height;

                // now have min terrain height on this volume
                // if volume entirely below, we can eliminate it
                if (minz >= originz + (z+1)*stepz) {
                    //cout << "deleted cube at " << xpos << ", " << ypos << ", " << zpos << endl;
                    ColoredVolume* cv = static_cast<ColoredVolume*>(*it);
                    // color
                    cv->r = cv->g = cv->b = cv->a = 0.0f;
                    if ((*it)->xn !=0) (*it)->xn->xp = 0;
                    if ((*it)->xp !=0) (*it)->xp->xn = 0;
                    if ((*it)->yn !=0) (*it)->yn->yp = 0;
                    if ((*it)->yp !=0) (*it)->yp->yn = 0;
                    if ((*it)->zn !=0) (*it)->zn->zp = 0;
                    if ((*it)->zp !=0) (*it)->zp->zn = 0;
                    //delete *it;
                    *it = 0;
                }
                // Otherwise, fill initialize color & position
                else {
                    ColoredVolume* cv = static_cast<ColoredVolume*>(*it);
                    // color
                    cv->r = cv->g = cv->b = cv->a = 0.0f;

/*                    float p = maxz + 0.5 * stepz;
                    if (p<zpos) p = zpos;
                    cv->position.set(xpos, ypos, p);
*/
                    // position: is terrain in volume, use offset from terrain floor
//                    if ((maxz > originz + z*stepz) || (minz < originz + (z+1)*stepz)) {
                    if (maxz >= originz + z*stepz) {
                        //cv->position.set(xpos, ypos, minz + 0.5 * stepz);
                        //cv->position.set(xpos, ypos, 100);
                        float p = 0.5 * (maxz + minz);
                        if (p > originz + (z+1)*stepz) p = originz + (z+1)*stepz;
                        if (p < originz + z*stepz) p = originz + z*stepz;
                        if (p<zpos) p = zpos;
                        cv->position.set(xpos, ypos, p);
                        //cout << "cube at " << xpos << ", " << ypos << ", " << minz + 0.5 * stepz << ", instead of " << zpos << endl;
                    }
                    // else use center of cube
                    else
                        cv->position.set(xpos, ypos, zpos);

                    // re-link valid cells only
                    if (prev!=0) prev->next = *it;
                    else firstVolume = *it;
                    prev = *it;
                }
                            

                ++it;
                xpos += stepx;
            }
            ypos += stepy;
        }
        zpos += stepz;
    }
    if (prev!=0) prev->next = 0;
    else firstVolume = 0;
    
}


void DrawableSporeCloud::step( )
{
    // process parent diffusion
    SporeCloud::step();

    // batch mode => no need for visible list!
    if (batch) return;
    
    ConcentrationList *cit;
    ColoredGenome* cg;
    
    // use local variables to avoid many de-references
    float r,g,b,c,a;

    // At this point, by parent pre-condition, we suppose no concurrent access
    // to the volume lists => we can safely run through them

    // There may be concurrent access on the colors, but this doesn't matter
    // if one frame draws a partially updated cloud.
    // On extreme case, there may be a flash of another color, though by continuity
    // of the diffusion reaction, this shouldn't be too visible in most cases (except when
    // a new genome comes in?)

    visibleListMutex.lock();
    ColoredVolume * vlistlast = 0;
    
    // Update colors
    for (ColoredVolume *cv = static_cast<ColoredVolume*>(firstProcess); cv != 0; cv = static_cast<ColoredVolume*>(cv->nextProcess)) {

        a = r = g = b = c = 0.0f;

        // For all genomes in this cell
        for (cit = cv->concentrations; cit!=0; cit = cit->next) {
            c = (float)cit->concentration;
            cg = (static_cast<ColoredGenome*>(cit->genome));
            r += cg->r * c;
            g += cg->g * c;
            b += cg->b * c;
            a += c;
        }
        
        // retain color only if it will be visible
        if (a>0.05f) {
            // end weighting of genome colors according to their concentration
            c = 1.0f / a;
            cv->r = r*c;
            cv->g = g*c;
            cv->b = b*c;
            
            // Don't create opaque clouds
            if (a>0.66f) a = 0.66f;
            cv->a = a;
            
            if (vlistlast==0) visibleList = cv;
            else vlistlast->nextVisible = cv;
            
            vlistlast = cv;

        }
        else cv->a = 0.0f;
    }

    // end list
    if (vlistlast==0) visibleList = 0;
    else vlistlast->nextVisible = 0;
    visibleListMutex.unlock();
    
}


#if defined(GL_VERTEX_PROGRAM_ARB)
/*static PFNGLGENPROGRAMSARBPROC           glGenProgramsARB;
static PFNGLDELETEPROGRAMSARBPROC        glDeleteProgramsARB;
static PFNGLBINDPROGRAMARBPROC           glBindProgramARB;
static PFNGLPROGRAMSTRINGARBPROC         glProgramStringARB;
static PFNGLPROGRAMENVPARAMETER4FARBPROC glProgramEnvParameter4fARB;
*/
static bool hasVertexProgram = false;
static bool testForVertexProgram = true;
//static GLuint vertex_program_ID[5];

static const char* vertex_program =
"!!ARBvp1.0\n"

"ATTRIB ipos = vertex.position;\n"
"ATTRIB icolor = vertex.color;\n"
"ATTRIB itex = vertex.texcoord[0];\n"

"PARAM mvp[4] = { state.matrix.mvp };\n"
"PARAM size = program.local[0];\n"

"OUTPUT opos = result.position;\n"
"OUTPUT ocolor = result.color;\n"
"OUTPUT otex = result.texcoord[0];\n"

// work registers
"TEMP tmp;\n"
"TEMP offset;\n"

// transform the position
"DP4 tmp.x, ipos, mvp[0];\n"
"DP4 tmp.y, ipos, mvp[1];\n"
"DP4 tmp.z, ipos, mvp[2];\n"
"DP4 tmp.w, ipos, mvp[3];\n"

// now according to tex position, we add offset in x/y for billboard
// 0.0 => -1 * width,  1.0 => +1*width, in both x and y
"ADD offset, itex, itex;\n"          // *2, in both x and y
"MAD offset, offset, size, -size;\n" // *size and -size, in both x and y

// output the final billboard position for this quad vertex
"ADD opos, tmp, offset;\n"

// transfer the texture coordinate & color
"MOV otex, itex;\n"
"MOV ocolor, icolor;\n"

"END\n"
// done!
;

#define MAY_HAVE_VERTEX_PROGRAM 1
#else
#define MAY_HAVE_VERTEX_PROGRAM 0
#endif


void DrawableSporeCloud::setDefaultAttributes( const std::string & texturefile, bool emissive_particles, bool lighting, int texture_unit )
{
    osg::StateSet *stateset = new osg::StateSet;

    stateset->setMode(GL_LIGHTING, lighting? osg::StateAttribute::ON: osg::StateAttribute::OFF);
    stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

    osg::Material *material = new osg::Material;
    material->setSpecular(osg::Material::FRONT, osg::Vec4(0, 0, 0, 1));
    material->setEmission(osg::Material::FRONT, osg::Vec4(0, 0, 0, 1));
    material->setColorMode(lighting? osg::Material::AMBIENT_AND_DIFFUSE : osg::Material::OFF);
    stateset->setAttributeAndModes(material, osg::StateAttribute::ON);

    if (!texturefile.empty()) {
        osg::Texture2D *texture = new osg::Texture2D;
        texture->setImage(osgDB::readImageFile(texturefile));
        texture->setFilter(osg::Texture2D::MIN_FILTER, osg::Texture2D::LINEAR);
        texture->setFilter(osg::Texture2D::MAG_FILTER, osg::Texture2D::LINEAR);
        stateset->setTextureAttributeAndModes(texture_unit, texture, osg::StateAttribute::ON);

        osg::TexEnv *texenv = new osg::TexEnv;
        texenv->setMode(osg::TexEnv::MODULATE);
        stateset->setTextureAttribute(texture_unit, texenv);
    }

    osg::BlendFunc *blend = new osg::BlendFunc;
    if (emissive_particles) {
        blend->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE);
    } else {
        blend->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE_MINUS_SRC_ALPHA);
    }
    stateset->setAttributeAndModes(blend, osg::StateAttribute::ON);

    // draw implementation without viewer???
    // consider Environment viewer not set correctly. Too bad!
    // We could also check only for current graphics context without OSG routines
/*    if ((!hasVertexProgram) && (viewer!=0)) {
        // check to see if vertex program is supported.
        // adapted from osgvertexprogram example program
        for(unsigned int contextID = 0; contextID < viewer->getDisplaySettings()->getMaxNumberOfGraphicsContexts(); ++contextID) {
            osg::VertexProgram::Extensions* vpExt = osg::VertexProgram::getExtensions(contextID,false);
            if (vpExt) {
                if (vpExt->isVertexProgramSupported()) {
                    hasVertexProgram = true;
                    break;
                } else {
                    hasVertexProgram = false;
                    break;
                }
            }
        }
        if (hasVertexProgram) {
        }
    }
*/    
    osg::VertexProgram* vp = new osg::VertexProgram();
    vp->setVertexProgram( vertex_program );
    vp->setProgramLocalParameter( 0, osg::Vec4( billboardWidth, billboardHeight, 0.0f, 0.0f ) );
    stateset->setAttributeAndModes( vp, osg::StateAttribute::ON | osg::StateAttribute::OVERRIDE );
    cout << "Using vertex program for cloud display." << endl;
    hasVertexProgram = true;

    setStateSet(stateset);

#if 0 //MAY_HAVE_VERTEX_PROGRAM

    if (testForVertexProgram) {
        testForVertexProgram = false;
        hasVertexProgram = false;

/*
#if (defined(_WIN32) || defined(WIN32))
        glGenProgramsARB           = (PFNGLGENPROGRAMSARBPROC)wglGetProcAddress("glGenProgramsARB");
        glDeleteProgramsARB        = (PFNGLDELETEPROGRAMSARBPROC)wglGetProcAddress("glDeleteProgramsARB");
        glBindProgramARB           = (PFNGLBINDPROGRAMARBPROC)wglGetProcAddress("glBindProgramARB");
        glProgramStringARB         = (PFNGLPROGRAMSTRINGARBPROC)wglGetProcAddress("glProgramStringARB");
        glProgramEnvParameter4fARB = (PFNGLPROGRAMENVPARAMETER4FARBPROC)wglGetProcAddress("glProgramEnvParameter4fARB");
#else
        glGenProgramsARB           = (PFNGLGENPROGRAMSARBPROC)glXGetProcAddress((const GLubyte *)"glGenProgramsARB");
        glDeleteProgramsARB        = (PFNGLDELETEPROGRAMSARBPROC)glXGetProcAddress((const GLubyte *)"glDeleteProgramsARB");
        glBindProgramARB           = (PFNGLBINDPROGRAMARBPROC)glXGetProcAddress((const GLubyte *)"glBindProgramARB");
        glProgramStringARB         = (PFNGLPROGRAMSTRINGARBPROC)glXGetProcAddress((const GLubyte *)"glProgramStringARB");
        glProgramEnvParameter4fARB = (PFNGLPROGRAMENVPARAMETER4FARBPROC)glXGetProcAddress((const GLubyte *)"glProgramEnvParameter4fARB");
#endif
*/
        glGenProgramsARB           = (PFNGLGENPROGRAMSARBPROC)osg::getGLExtensionFuncPtr("glGenProgramsARB");
        glDeleteProgramsARB        = (PFNGLDELETEPROGRAMSARBPROC)osg::getGLExtensionFuncPtr("glDeleteProgramsARB");
        glBindProgramARB           = (PFNGLBINDPROGRAMARBPROC)osg::getGLExtensionFuncPtr("glBindProgramARB");
        glProgramStringARB         = (PFNGLPROGRAMSTRINGARBPROC)osg::getGLExtensionFuncPtr("glProgramStringARB");
        glProgramEnvParameter4fARB = (PFNGLPROGRAMENVPARAMETER4FARBPROC)osg::getGLExtensionFuncPtr("glProgramEnvParameter4fARB");


        if( (glGenProgramsARB!=0) && (glDeleteProgramsARB!=0) && (glBindProgramARB!=0) && (glProgramStringARB!=0) && (glProgramEnvParameter4fARB!=0)) {

            // reset error
            glGetError();
            
            glGenProgramsARB(1, vertex_program_ID);
            if ( glGetError() != GL_NO_ERROR ) {
                cout << "Error allocating vertex program ID" << endl;
                return;
            }
            glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vertex_program_ID[0] );
            if ( glGetError() != GL_NO_ERROR ) {
                cout << "Error binding vertex program ID" << endl;
                return;
            }
            glProgramStringARB( GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(vertex_program), vertex_program);

            if ( glGetError() == GL_NO_ERROR ) {
                glProgramEnvParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, halfBillboardWidth, halfBillboardWidth, 0, 0);
                if ( glGetError() != GL_NO_ERROR ) {
                    cout << "Error setting vertex program parameter." << endl;
                    return;
                }
                hasVertexProgram = true;
                cout << "Using vertex program for cloud display." << endl;

            } else {
                GLint pos;
                glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &pos);
                cerr << "Vertex program error at position " << pos << ": " << (char*)glGetString(GL_PROGRAM_ERROR_STRING_ARB) << endl;
            }

        }

    }
#endif

}


void DrawableSporeCloud::drawImplementation(osg::State &state) const
{

    if (viewDisabled) return;

#if MAY_HAVE_VERTEX_PROGRAM
    // Do the matrix tranform and billboard offset on the GPU if possible
    if (hasVertexProgram) {
        glPushAttrib(GL_DEPTH_BUFFER_BIT);
        glDepthMask(GL_FALSE);
        
/*        
        glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vertex_program_ID[0] );
        if ( glGetError() != GL_NO_ERROR ) cout << "error bind" <<endl;

        glProgramEnvParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, halfBillboardWidth, halfBillboardWidth, 0, 0);
        if ( glGetError() != GL_NO_ERROR ) cout << "error set param" <<endl;
*/
//        glEnable( GL_VERTEX_PROGRAM_ARB );
        glBegin(GL_QUADS);

        float* f;
        visibleListMutex.lock();
        // Main bottleneck is fetching elements from main memory, thanks oprofile
        // in this example, fetching cv->a to test if !=0.0f takes 3-4% of program!!!
        // So, really, fetching only the necessary volumes with visibility list is good!
        for (ColoredVolume *cv = visibleList; cv != 0; cv = cv->nextVisible) {
            glColor4f( cv->r, cv->g, cv->b, cv->a);
            f = cv->position.ptr();
            glTexCoord2f(0.0f, 0.0f);
            glVertex3fv(f); // offset on the GPU
            glTexCoord2f(1.0f, 0.0f);
            glVertex3fv(f); // offset on the GPU
            glTexCoord2f(1.0f, 1.0f);
            glVertex3fv(f); // offset on the GPU
            glTexCoord2f(0.0f, 1.0f);
            glVertex3fv(f); // offset on the GPU
        }
        visibleListMutex.unlock();
    
        glEnd();

        glFlush();

//        glDisable( GL_VERTEX_PROGRAM_ARB );

        // restore depth mask settings
        glPopAttrib();
        return;   
    }


#endif

    // Definitely no vertex program. Do it on CPU

    // get the current modelview matrix
    osg::Matrix modelview = state.getModelViewMatrix();
    state.applyModelViewMatrix(0);

    // set up depth mask for first rendering pass
    glPushAttrib(GL_DEPTH_BUFFER_BIT);
    glDepthMask(GL_FALSE);

    // render, first pass
    single_pass_render(modelview);

    // restore depth mask settings
    glPopAttrib();

/*        // render, second pass
    if (doublepass_) {
        // set up color mask for second rendering pass
        glPushAttrib(GL_COLOR_BUFFER_BIT);
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

        // render the particles onto the depth buffer
        single_pass_render(state, modelview);

        // restore color mask settings
        glPopAttrib();
    }
*/
}


void DrawableSporeCloud::single_pass_render(const osg::Matrix &modelview) const
{
    glBegin(GL_QUADS);

    visibleListMutex.lock();
    // Main bottleneck is fetching elements from main memory, thanks oprofile
    // in this example, fetching cv->a to test if !=0.0f takes 3-4% of program!!!
    // So, really, fetching only the necessary volumes with visibility list is good!
    for (ColoredVolume *cv = visibleList; cv != 0; cv = cv->nextVisible) {
        glColor4f( cv->r, cv->g, cv->b, cv->a);
        renderParticle(modelview.preMult(cv->position));
    }
    visibleListMutex.unlock();

    glEnd();

}

void DrawableSporeCloud::renderParticle(const osg::Vec3 &pos) const
{
    float f[3];
    f[0] = pos.x() - halfBillboardWidth;
    f[1] = pos.y() - halfBillboardHeight;
    f[2] = pos.z();
    glTexCoord2f(0.0f, 0.0f);
    glVertex3fv(f);
    f[0] += billboardWidth;
    glTexCoord2f(1.0f, 0.0f);
    glVertex3fv(f);
    f[1] += billboardHeight;
    glTexCoord2f(1.0f, 1.0f);
    glVertex3fv(f);
    f[0] -= billboardWidth;
    glTexCoord2f(0.0f, 1.0f);
    glVertex3fv(f);
}

osg::Object * DrawableSporeCloud::cloneType( ) const
{
    // Ignore OSG stuff for now
    return const_cast<osg::Object*>(static_cast<const osg::Object*>(this));
}

osg::Object * DrawableSporeCloud::clone( const osg::CopyOp & copyop ) const
{
    // Ignore OSG stuff for now
    return const_cast<osg::Object*>(static_cast<const osg::Object*>(this));
}

/*
void DrawableSporeCloud::lock(SporeVolume* volume) const
{
    reinterpret_cast<OpenThreads::Mutex*>(volume->userObject)->lock();
}

void DrawableSporeCloud::unlock(SporeVolume* volume) const
{
    reinterpret_cast<OpenThreads::Mutex*>(volume->userObject)->unlock();
}

*/
