/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DRAWABLESPORECLOUD_H
#define DRAWABLESPORECLOUD_H

#include <OpenThreads/Mutex>

#include <osg/Drawable>

#include "Terrain.h"
#include "SporeCloud.h"

/**
@author Nicolas Brodu
*/
class DrawableSporeCloud : public osg::Drawable, public SporeCloud
{
public:
    DrawableSporeCloud(Terrain* terrain, int divx, int divy, int divz, double baseDiffusionCoef, double lowestConcentration);
             

    /// Taken from OSG particle system. Used to set up a texture for particles
    void setDefaultAttributes(const std::string &texturefile="", bool emissive_particles=true, bool lighting=false, int texture_unit=0);

    // Drawable
    void drawImplementation(osg::State &state) const;
    
    struct ColoredGenome : public SporeCloud::Genome {
        float r,g,b;
        inline ColoredGenome() : r(0.0f), g(0.0f), b(1.0f) {}
    };

    // just in case default not OK
    void setBillboardSize(float width, float height) {
        billboardWidth = width;
        billboardHeight = height;
        halfBillboardWidth = 0.5 * width;
        halfBillboardHeight = 0.5 * height;
    }

    /// Override parent integration step to compute visual updates at this
    /// frequency, and not each frame!
    virtual void step();

    /** Whether to display the cloud, or not. Default is true,
        but when the cloud is too dense it may be de-activated
        so we can see the agents.
    */
    inline void setEnableView(bool state) {viewDisabled = !state;}
    inline bool getEnableView() {return !viewDisabled;}
    
protected:
    // does the actual rendering
    void single_pass_render(const osg::Matrix &modelview) const;
    void renderParticle(const osg::Vec3 &pos) const;
    
    Terrain* terrain; // for min Z
    double baseDiffusionCoef; // useless, for OSG object stuff only
    
    float billboardWidth, billboardHeight;
    float halfBillboardWidth, halfBillboardHeight;

    // filled up at init time from div and size
    double stepx, stepy, stepz;

    struct ColoredVolume : public SporeVolume {
        float r, g, b, a;
        // trade off memory for speed => cache position
        osg::Vec3 position;
        ColoredVolume* nextVisible;
    };
    static SporeVolume* coloredVolumeFactory(void* user);

    ColoredVolume* visibleList;
    mutable OpenThreads::Mutex visibleListMutex;
    bool viewDisabled;

public:
    // OSG object stuff
    virtual osg::Object* cloneType() const;
    virtual osg::Object* clone(const osg::CopyOp& copyop) const;
    virtual bool isSameKindAs(const osg::Object* obj) const { return dynamic_cast<const DrawableSporeCloud*>(obj)!=NULL; }
    virtual const char* libraryName() const { return "crogai"; }
    virtual const char* className() const { return "DrawableSporeCloud"; }
};

#endif
