/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include "Environment.h"

namespace Environment {

    double cycleX = 0.0;
    double cycleY = 0.0;
    double cycleOffsetX = 0.0;
    double cycleOffsetY = 0.0;

    double dragCoefficient = 0.0;
    double gravity = 0.0;

    Simulator* simu = 0;
    EventQueueViewer* viewer = 0;
    DrawableSporeCloud* cloud = 0;
    osg::Group* root = 0;
    
    double agentsAITimeStep = 0.15;
    double agentsIntegrationTimeStep = 0.04;
    double cloudIntegrationTimeStep = 0.3;
    
    bool batch = false;

    // default values
    int WindowWidth = 800;
    int WindowHeight = 600;
    int WindowXOffset = 240;
    int WindowYOffset = 212;

    
    const double wrappedDist2(const osg::Vec3d& p1, const osg::Vec3d& p2) {
        const double* p1ptr = p1.ptr();
        const double* p2ptr = p2.ptr();
        if (cycleX==0.0) {
            if (cycleY==0.0) {
                return square(p1ptr[0] - p2ptr[0])
                    + square(p1ptr[1] - p2ptr[1])
                    + square(p1ptr[2] - p2ptr[2]);
            } else {
                return square(p1ptr[0] - p2ptr[0])
                    + square(remainder(p1ptr[1] - p2ptr[1], cycleY))
                    + square(p1ptr[2] - p2ptr[2]);
            }
        } else {
            if (cycleY==0.0) {
                return square(remainder(p1ptr[0] - p2ptr[0], cycleX))
                    + square(p1ptr[1] - p2ptr[1])
                    + square(p1ptr[2] - p2ptr[2]);
            } else {
                return square(remainder(p1ptr[0] - p2ptr[0], cycleX))
                    + square(remainder(p1ptr[1] - p2ptr[1], cycleY))
                    + square(p1ptr[2] - p2ptr[2]);
            }
        }
    }
    
    // For private use in this compilation unit
    inline const double wrapCoord(double coord, const double& offset, const double& cycle) {
        coord = fmod(coord - offset, cycle);
        if (coord<0) coord += cycle;
        return coord + offset;
    }
    
    const osg::Vec3d wrapPosition(const osg::Vec3d& position) {
        const double* ptr = position.ptr();
        if (cycleX==0.0) {
            if (cycleY==0.0) {
                return position;
            } else {
                return osg::Vec3d(ptr[0],
                                  wrapCoord(ptr[1], cycleOffsetY, cycleY),
                                  ptr[2]);
            }
        } else {
            if (cycleY==0.0) {
                return osg::Vec3d(wrapCoord(ptr[0], cycleOffsetX, cycleX),
                                  ptr[1],
                                  ptr[2]);
            } else {
                return osg::Vec3d(wrapCoord(ptr[0], cycleOffsetX, cycleX),
                                  wrapCoord(ptr[1], cycleOffsetY, cycleY),
                                  ptr[2]);
            }
        }
    }

    const osg::Vec3d wrappedDiffVector(const osg::Vec3d& p1, const osg::Vec3d& p2) {
        const double* p1ptr = p1.ptr();
        const double* p2ptr = p2.ptr();
        if (cycleX==0.0) {
            if (cycleY==0.0) {
                return osg::Vec3d(p1ptr[0] - p2ptr[0],
                                p1ptr[1] - p2ptr[1],
                                p1ptr[2] - p2ptr[2]);
            } else {
                return osg::Vec3d(p1ptr[0] - p2ptr[0],
                                remainder(p1ptr[1] - p2ptr[1], cycleY),
                                p1ptr[2] - p2ptr[2]);
            }
        } else {
            if (Environment::cycleY==0.0) {
                return osg::Vec3d(remainder(p1ptr[0] - p2ptr[0], cycleX),
                                p1ptr[1] - p2ptr[1],
                                p1ptr[2] - p2ptr[2]);
            } else {
                return osg::Vec3d(remainder(p1ptr[0] - p2ptr[0], cycleX),
                                remainder(p1ptr[1] - p2ptr[1], cycleY),
                                p1ptr[2] - p2ptr[2]);
            }
        }
    }

}
