/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef LOGGER_H
#define LOGGER_H

#include <vector>
#include <string>

/**
    Very basic logging interface.
    Enough for our current needs, something more sophisticated may be necessary in the future.
    See for example the Transport Sample Protocol at https://savannah.nongnu.org/projects/tsp/

    The idea is the use the logger as a vector of values. When all new values are filled,
    then call record() to write to the log file/socket/whatever.
    The logger should be initialized with a descriptive ID (used in file name, for ex:), and
    the variables names corresponding to the values that will be sent.
    
    @author Nicolas Brodu
*/
struct Logger : public std::vector<double> {
    virtual ~Logger();
    std::vector<std::string> names;
    std::string id;

    // Reuse iterators from parent. Implicit typenames have been deprecated in C++
    typedef std::vector<double>::const_iterator const_iterator;
    typedef std::vector<double>::iterator iterator;

    /// Takes a number of entries and their names
    /// This implicitly defines this vector size.
    /// All values are initialized with 0.0
    Logger(const std::string& id, const std::vector<std::string>& names);

    /// Record the previously set values. If a value was not set since the last
    /// record, it keeps the same old value (init:0.0) for the new one.
    virtual void record();
};

#endif
