/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef OBSTACLE_H
#define OBSTACLE_H

#include <osg/Group>

#include "TranslateTransform.h"
#include "Localized.h"

/**

    Obstacle is a Localized Object with fixed position.
    If using Bricks and a locality DB, it shold be created before the obstacles.
    The obstacles automatically add themselves in the DB at their position if it is in use.
    
    Its avatar should be built once and inserted in the scene graph,
    and cannot move later on.

@author Nicolas Brodu
*/
class Obstacle : public Localized
{
public:
    
    Obstacle(const osg::Vec3d& pos);
    Obstacle(const double& x, const double& y, const double& z);
    
    /** Add this obstacle avatar to the scene graph

        External API. Subclasses should override buildAvatar and position their
        model at 0,0,0 with the correct orientation.
        
        A translation transform is inserted to handle positioning
        in this higher-level class
        
        Warning: caller responsibility to call this function at the right
        moment. the delay parameter may be used to postpone addition of
        the avatar to the next viewer update
    */
    virtual void addAvatarToGroup(osg::Group* group, bool delayToNextFrame = false);

    /** Returns the osg Node for this obstacle.
        Translation of the agent's avatar at its position.
     */
    inline osg::Node* getNode() {return transform;}
    
    /** Returns the osg Node radius for this obstacle. commodity for avoiding recomputing
        from obstacle's node each step
     */
    inline double getRadius() {return radius;}

    /** Artifact for schedulable method: void return type */
    virtual void updateNeighbors(double radius);
    
protected:
    
    TranslateTransform *transform;
    double radius;

    /** Build this agent avatar. Default is a white sphere of radius 1.
        For subclasses to override, the model should be created at 0,0,0 with the correct orientation.
    */
    virtual osg::Node* buildAvatar();
    
};

#endif
