/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <sstream>
#include <fstream>

#include <OpenThreads/Mutex>
#include <OpenThreads/Thread>

#include <osg/Array>
#include <osg/PrimitiveSet>
#include <osg/Geometry>
#include <osg/Geode>

#include <osg/BlendFunc>
#include <osg/Projection>
#include <osg/MatrixTransform>

#include <osgGA/GUIEventHandler>
#include <osgGA/TrackballManipulator>

#include "CrogaiAgent.h"
#include "Terrain.h"
#include "TranslateTransform.h"
#include "AgentCore.h"
#include "DrawableSporeCloud.h"

#include "Utility.h"
#include "Environment.h"


using namespace Environment;
using namespace std;

#include <math.h>
#include <assert.h>
#include <stdio.h>

#include <time.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*

==============================================================================

The following class definition is taken from "randomc.h", a C++ random number
generator library released uner the GNU GPL by Agner Fog.

See http://www.agner.org/random/randomc.htm for more information and the
original files.

This derived work contains only the definition of the Mersenne twister
random number generator, and was adapted to:
- Remove unused constants
- TODO: Made architecture detection from configure script
*/

// Define 32 bit signed and unsigned integers.
// Change these definitions, if necessary, on 64 bit computers
typedef   signed long int32;     
typedef unsigned long uint32;     

class TRandomMersenne {                // encapsulate random number generator
    // constants for MT19937:
    #define MERS_N   624
    #define MERS_M   397
    #define MERS_R   31
    #define MERS_U   11
    #define MERS_S   7
    #define MERS_T   15
    #define MERS_L   18
    #define MERS_A   0x9908B0DF
    #define MERS_B   0x9D2C5680
    #define MERS_C   0xEFC60000
public:
    TRandomMersenne(uint32 seed) {       // constructor
        RandomInit(seed);}

    void RandomInit(uint32 seed);        // re-seed
    void RandomInitByArray(uint32 seeds[], int length); // seed by more than 32 bits
    int IRandom(int min, int max);       // output random integer
    double Random();                     // output random float
    uint32 BRandom();                    // output random bits

private:
    uint32 mt[MERS_N];                   // state vector
    int mti;                             // index into mt

// win32 problem: defines in winsock2.h prevent correct enum
// unconditional undef for this cpp file as we don't use these constants here anyway
#undef LITTLEENDIAN
#undef BIGENDIAN
#undef NONIEEE
    
    enum TArch {LITTLEENDIAN, BIGENDIAN, NONIEEE};
    TArch Architecture;                  // conversion to float depends on computer architecture
};    

#include "mersenne.cpp"

/* 
End of randomc library re-use
==============================================================================
*/

namespace Utility {

osgGA::TrackballManipulator *trackBallManipulator = 0;

EventQueueViewer* createViewer( const char * windowTitle ) {
    
    // create a new render surface
    Producer::RenderSurface *rsWindow = new Producer::RenderSurface;
    
    // set the number of screens
    rsWindow->setScreenNum(0);
    
    // set the window title bar caption
    rsWindow->setWindowName(windowTitle);
    
    // set the window client size
    rsWindow->setWindowRectangle(WindowXOffset,WindowYOffset,WindowWidth,WindowHeight);
    
    // create a new camera
    Producer::Camera *camera1 = new Producer::Camera;
    
    // set the render surface of the camera to the window that was just created
    camera1->setRenderSurface(rsWindow);

    // set the size of the camera viewport to the same size as the window
    camera1->setProjectionRectangle(WindowXOffset,WindowYOffset,WindowWidth,WindowHeight);
    
    // create a new camera configuration container
    Producer::CameraConfig *cfg = new Producer::CameraConfig;
    
    // add the camera that was just created to the configuration
    cfg->addCamera("Camera One",camera1);

    // Finally create the viewer
    viewer = new EventQueueViewer(cfg);

    //viewer->setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);

    trackBallManipulator = new osgGA::TrackballManipulator;
    
    // Set Up an OSG viewer with only one manipulator
    viewer->setUpViewer(osgProducer::Viewer::ESCAPE_SETS_DONE|osgProducer::Viewer::SKY_LIGHT_SOURCE|osgProducer::Viewer::HEAD_LIGHT_SOURCE|osgProducer::Viewer::VIEWER_MANIPULATOR);
    viewer->selectCameraManipulator(viewer->addCameraManipulator(trackBallManipulator));
}



osg::Group *drawLine(osg::Vec3d p1, osg::Vec3d p2, osg::Vec4 c1, osg::Group *lineGroup) {
    drawLine(p1, p2, c1, c1, lineGroup);
}

osg::Group *drawLine(osg::Vec3d p1, osg::Vec3d p2, osg::Vec4 c1, osg::Vec4 c2, osg::Group *lineGroup) {
    
    osg::Vec3Array* vertices = new osg::Vec3Array;
    vertices->push_back( p1 );
    vertices->push_back( p2 );
    osg::Geometry* lineGeometry = new osg::Geometry();
    lineGeometry->setVertexArray( vertices );

    osg::DrawElementsUInt* lineElements = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
    lineElements->push_back(0);
    lineElements->push_back(1);
    lineGeometry->addPrimitiveSet(lineElements);
    
    osg::Vec4Array* colors = new osg::Vec4Array;
    colors->push_back( c1 ); //index 0 green
    colors->push_back( c2 ); //index 1 red

    osg::TemplateIndexArray<unsigned int, osg::Array::UIntArrayType,4,4> *colorIndexArray;
    colorIndexArray = new osg::TemplateIndexArray<unsigned int, osg::Array::UIntArrayType,4,4>;
    colorIndexArray->push_back(0); // vertex 0 assigned color array element 0
    colorIndexArray->push_back(1); // vertex 1 assigned color array element 1

    lineGeometry->setColorArray(colors);
    lineGeometry->setColorIndices(colorIndexArray);
    lineGeometry->setColorBinding(osg::Geometry::BIND_PER_VERTEX);

    lineGeometry->setUseDisplayList( false );
    //if (lineNode!=0) transform->getParent(0)->removeChild(lineNode);
    osg::Geode* lineNode =new osg::Geode();
    lineNode->getOrCreateStateSet()->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
    lineNode->addDrawable(lineGeometry);

    if (lineGroup==0) lineGroup = new osg::Group();
    lineGroup->addChild(lineNode);
}

// Use unique instances as we use the unique viewer anyway...
static osg::Geode* topHUD = 0;
static osg::Geode* bottomHUD = 0;

static void addSceneHelper(osg::Projection* p) {
    osg::Group* topmost = dynamic_cast<osg::Group*>(viewer->getTopMostSceneData());
    if (topmost==0) return;
    topmost->addChild(p);
}

/** GRRRRRRRRRRRRRRRRRR!!!!! OSG didn't make setText virtual!!!

    We might as well have created another class...
    Anyway, so long as it's only used internally, we known what we're doing...
*/
struct EventFusionnedText : public osgText::Text {

    OpenThreads::Mutex access;
    std::string nextText;
    bool updateFlag;

    EventFusionnedText() {
        updateFlag = false;
    }

    // set text locally, and update parent string only once per frame, on demand
    void setText (const std::string &text) {
        access.lock();
        nextText = text;
        if ((!updateFlag) && (viewer!=0))
            viewer->post(this, &EventFusionnedText::update);
        updateFlag = true;
        access.unlock();
    }

    void update() {
        access.lock();
        osgText::Text::setText(nextText);
        updateFlag = false;
        access.unlock();
    }
        
};

static EventFusionnedText* topLeftText = 0;
static EventFusionnedText* topRightText = 0;
static EventFusionnedText* bottomLeftText = 0;
static EventFusionnedText* bottomRightText = 0;

static EventFusionnedText* getTextHUD(TextHUDPosition where) {

    // batch mode, or graphics not yet set up
    if (batch || (viewer==0)) return 0;

    osg::Geode** hud = 0;
    switch (where) {
        case TOP_LEFT:
            if (topLeftText!=0) return topLeftText;
            hud = &topHUD;
            break;
        case TOP_RIGHT:
            if (topRightText!=0) return topRightText;
            hud = &topHUD;
            break;
        case BOTTOM_LEFT:
            if (bottomLeftText!=0) return bottomLeftText;
            hud = &bottomHUD;
            break;
        case BOTTOM_RIGHT:
            if (bottomRightText!=0) return bottomRightText;
            hud = &bottomHUD;
            break;
    };

    // check if we need to create the HUD or re-use it
    if (*hud==0) {
        osg::Group* topmost = dynamic_cast<osg::Group*>(viewer->getTopMostSceneData());
        if (topmost==0) return 0;
        
        *hud = new osg::Geode();
        
        // turn lighting off for the text and disable depth test to ensure its always ontop.
        osg::StateSet* stateset = (*hud)->getOrCreateStateSet();
        stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
    
        // Disable depth test, and make sure that the hud is drawn after everything
        // else so that it always appears ontop.
        stateset->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
        stateset->setRenderBinDetails(11,"RenderBin");

        osg::BoundingBox bb;
        switch (where) {
            case TOP_LEFT:
            case TOP_RIGHT:
                bb = osg::BoundingBox(osg::Vec3(0,WindowHeight-WindowHeight/25.0-WindowHeight/100.0,0),osg::Vec3(WindowWidth,WindowHeight,0));
                break;
            case BOTTOM_LEFT:
            case BOTTOM_RIGHT:
                bb = osg::BoundingBox(osg::Vec3(0,0,0),osg::Vec3(WindowWidth,WindowHeight/25.0+WindowHeight/100.0,0));
                break;
        };
        osg::Geometry* geom = new osg::Geometry;
    
        osg::Vec3Array* vertices = new osg::Vec3Array;
        float depth = bb.zMin()-0.1;
        vertices->push_back(osg::Vec3(bb.xMin(),bb.yMax(),depth));
        vertices->push_back(osg::Vec3(bb.xMin(),bb.yMin(),depth));
        vertices->push_back(osg::Vec3(bb.xMax(),bb.yMin(),depth));
        vertices->push_back(osg::Vec3(bb.xMax(),bb.yMax(),depth));
        geom->setVertexArray(vertices);
    
        osg::Vec3Array* normals = new osg::Vec3Array;
        normals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
        geom->setNormalArray(normals);
        geom->setNormalBinding(osg::Geometry::BIND_OVERALL);
    
        osg::Vec4Array* colors = new osg::Vec4Array;
        colors->push_back(osg::Vec4(1.0f,1.0,0.8f,0.2f));
        geom->setColorArray(colors);
        geom->setColorBinding(osg::Geometry::BIND_OVERALL);
    
        geom->addPrimitiveSet(new osg::DrawArrays(GL_QUADS,0,4));
        
        osg::StateSet* stateset2 = geom->getOrCreateStateSet();
        stateset2->setMode(GL_BLEND,osg::StateAttribute::ON);
        //stateset2->setAttribute(new osg::PolygonOffset(1.0f,1.0f),osg::StateAttribute::ON);
        stateset2->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
    
        (*hud)->addDrawable(geom);
    
        // create the hud.
        osg::MatrixTransform* modelview_abs = new osg::MatrixTransform;
        modelview_abs->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
        modelview_abs->setMatrix(osg::Matrix::identity());
        modelview_abs->addChild(*hud);
    
        osg::Projection* projection = new osg::Projection;
        projection->setMatrix(osg::Matrix::ortho2D(0,WindowWidth,0,WindowHeight));
        projection->addChild(modelview_abs);

        viewer->post(&addSceneHelper, projection);
//        topmost->addChild(projection);
    }
    
    EventFusionnedText* theText = new EventFusionnedText;
    (*hud)->addDrawable( theText );
    theText->setFont("FreeSans.ttf");
    theText->setText("");
    //scoresText->setCharacterSizeMode(osgText::SCREEN_COORDS);
    theText->setCharacterSize(WindowHeight/25.0);
    
    switch (where) {
        case TOP_LEFT:
            topLeftText = theText;
            theText->setPosition(osg::Vec3(WindowWidth/1000.0,WindowHeight-WindowHeight/100.0,0.0f));
            theText->setAlignment(osgText::Text::LEFT_TOP);
            break;
        case TOP_RIGHT:
            topRightText = theText;
            theText->setPosition(osg::Vec3(WindowWidth - WindowWidth/1000.0,WindowHeight-WindowHeight/100.0,0.0f));
            theText->setAlignment(osgText::Text::RIGHT_TOP);
            break;
        case BOTTOM_LEFT:
            bottomLeftText = theText;
            theText->setPosition(osg::Vec3(WindowWidth/1000.0,WindowHeight/200.0,0.0f));
            theText->setAlignment(osgText::Text::LEFT_BOTTOM);
            break;
        case BOTTOM_RIGHT:
            bottomRightText = theText;
            theText->setPosition(osg::Vec3(WindowWidth - WindowWidth/1000.0,WindowHeight/200.0,0.0f));
            theText->setAlignment(osgText::Text::RIGHT_BOTTOM);
            break;
    };

    return theText;
}

bool statusToConsole = false;
bool statusTimeStamped = false;

/** Display some status text. Creates a text HUD if necessary */
void setStatus(const std::string& status, TextHUDPosition where, bool alsoOut) {

    // create or reuse
    EventFusionnedText* statusText = getTextHUD(where);

    if (statusTimeStamped) {
        std::string status2;
        ostringstream os;
        os << "t=" << simu->getTime() << ": " << status;
        status2 = os.str();

        // batch mode, graphics not yet set up, or explicit demand
        if ((statusText==0) || alsoOut) cout << status2 << endl;

        if (statusText!=0) statusText->setText(status2);
        return;
    }

    // batch mode, graphics not yet set up, or explicit demand
    if ((statusText==0) || alsoOut) cout << status << endl;
    
    if (statusText!=0) statusText->setText(status);
}

static bool firstTime = true;
static double sysTimeOri = 0.0;
static double simTimeOri = 0.0;
// This member allows to cancel timer
static const Simulator::GenericDispatcher* checkTimeEvent = 0;

void checkTime() {
    if (firstTime) {
        firstTime = false;
        sysTimeOri = simu->getSystemTime();
        simTimeOri = simu->getTime();
        return;
    }
    double sysTime = simu->getSystemTime() - sysTimeOri;
    double simTime = simu->getTime() - simTimeOri;
    ostringstream os;
    os << "Time advance: real:" << sysTime << ", simu:" << simTime << ", ratio=" << simTime / sysTime << endl;
    Utility::setStatus(os.str());
}

static const Simulator::GenericDispatcher* displayTimeEvent = 0;

void displayTime() {
    double simTime = simu->getTime() - simTimeOri;
    ostringstream os;
    os << "Time: " << simu->getTime() << endl;
    Utility::setStatus(os.str(), Utility::BOTTOM_RIGHT);
}


void toggleSimulationTime() {
    if (simu==0) return;
    if (displayTimeEvent!=0) {
        simu->cancel(displayTimeEvent);
        setStatus("", Utility::BOTTOM_RIGHT);
        displayTimeEvent = 0;
    }
    else {
        // display each exact second of simulator time: schedule for exact boundary
        displayTimeEvent = simu->postCyclicAt((int)(simu->getTime()+1.0), 1.0, &displayTime);
    }
}

void resetView() {
    osgGA::GUIEventAdapter* ea = new osgGA::GUIEventAdapter();
    if (trackBallManipulator) trackBallManipulator->home(*ea, *viewer);
    ea->unref();
}

static long randomSeed = 42;

class KeyboardEventHandler : public osgGA::GUIEventHandler {
public:

    KeyboardEventHandler(osg::Group* tileableNode, osg::Group* tileableParent) {
        root = tileableNode;
        metaroot = tileableParent;
        normalGroup = 0;
    }

    osg::Group *root, *metaroot, *normalGroup;

    inline void translateRoot(int tx, int ty) {
        TranslateTransform* tt=0;
        tt = new TranslateTransform(tx*cycleX, ty*cycleY, 0);
        tt->addChild(root);
        metaroot->addChild(tt);
    }

    virtual bool handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter&) {
        switch(ea.getEventType()) {

            case osgGA::GUIEventAdapter::KEYDOWN:
                switch(ea.getKey()) {
                    case 'g': {
                        gravity += 0.1;
                        ostringstream os;
                        os << "gravity is now " << gravity;
                        setStatus(os.str());
                        } break;
                    
                    case 'G': {
                        gravity -= 0.1;
                        if (gravity<0) gravity=0;
                        ostringstream os;
                        os << "gravity is now " << gravity;
                        setStatus(os.str());
                        } break;
                    
                    case ' ':
                        if (simu->isRunning()) simu->stop();
                        else simu->start();
                        break;

                    case 'n':
                    case 'N': {
                        Terrain* terrain = dynamic_cast<Terrain*>(TerrainAgent::terrain);
                        if (terrain==0) break;
                        if (normalGroup==0) {
                            // OSG scene graph operations OK in event handler
                            normalGroup = terrain->drawNormals();
                            root->addChild(normalGroup);
                            setStatus("Floor normals on");
                        }
                        else {
                            // OSG scene graph operations OK in event handler
                            root->removeChild(normalGroup);
                            normalGroup = 0;
                            setStatus("Floor normals off");
                        }
                        } break;

                    case 'C':
                    case 'c':
                        if (metaroot==0) break;
                        // keep only the first child, root
                        if (metaroot->getNumChildren() > 9) {
                            metaroot->removeChild(1, metaroot->getNumChildren() -1);
                            setStatus("Cyclic terrain visualization off");
                            break;
                        }
                        if (metaroot->getNumChildren() >1 ) {
                            translateRoot(-2,-2);
                            translateRoot(-2,-1);
                            translateRoot(-2,0);
                            translateRoot(-2,1);
                            translateRoot(-2,2);
                            translateRoot(2,-2);
                            translateRoot(2,-1);
                            translateRoot(2,0);
                            translateRoot(2,1);
                            translateRoot(2,2);
                            translateRoot(-1,-2);
                            translateRoot(0,-2);
                            translateRoot(1,-2);
                            translateRoot(-1,2);
                            translateRoot(0,2);
                            translateRoot(1,2);
                            setStatus("Cyclic terrain visualization: level 2");
                            break;
                        }
                        translateRoot(1,0);
                        translateRoot(-1,0);
                        translateRoot(0,1);
                        translateRoot(0,-1);
                        translateRoot(1,1);
                        translateRoot(-1,1);
                        translateRoot(1,-1);
                        translateRoot(-1,-1);

                        setStatus("Cyclic terrain visualization: level 1");
                        break;

                    case 'm':
                    case 'M':
                        if (simu==0) break;
                        if (checkTimeEvent!=0) {
                            simu->cancel(checkTimeEvent);
                            setStatus("Performance timer stopped");
                            checkTimeEvent = 0;
                        }
                        else {
                            setStatus("Initializing performance timer");
                            firstTime = true;
                            checkTimeEvent = simu->postCyclic(5.0, &checkTime);
                        }
                        break;
                    
                    case osgGA::GUIEventAdapter::KEY_BackSpace:
                        setStatus("");
                        break;

                    case 'v':
                    case 'V':
                        resetView();
                        break;
                    
                    case 'r':
                    case 'R': {
                        ostringstream os;
                        os << "Random seed: " << randomSeed;
                        setStatus(os.str());
                        } break;

                    case 's':
                    case 'S':
                        if (cloud) cloud->setEnableView(!cloud->getEnableView());
                        break;
                    
                    case 't':
                    case 'T':
                        toggleSimulationTime();
                        break;
                    
                    case 'p':
                    case 'P':
                        CrogaiAgent::togglePopulationGraph();
                        break;
    
/*                    case '+': {
                        if (!simu) break;
                        double tf = simu->getTimeFactor() + 0.1;
                        simu->setTimeFactor(tf);
                        ostringstream os;
                        os << "Simulator now targetting " << tf << " time ratio.";
                        setStatus(os.str());
                        } break;
                        
                    case '-': {
                        if (!simu) break;
                        double tf = simu->getTimeFactor() - 0.1;
                        if (tf<0.1) tf = 0.1;
                        simu->setTimeFactor(tf);
                        ostringstream os;
                        os << "Simulator now targetting " << tf << " time ratio.";
                        setStatus(os.str());
                        } break;
                    
                    case '*': {
                        if (!simu) break;
                        double tf = simu->getTimeFactor() * 2.0;
                        simu->setTimeFactor(tf);
                        ostringstream os;
                        os << "Simulator now targetting " << tf << " time ratio.";
                        setStatus(os.str());
                        } break;
                        
                    case '/': {
                        if (!simu) break;
                        double tf = simu->getTimeFactor() * 0.5;
                        simu->setTimeFactor(tf);
                        ostringstream os;
                        os << "Simulator now targetting " << tf << " time ratio.";
                        setStatus(os.str());
                        } break;
*/                    
                    default:
                        return false;
                }
                return true;
/*
            case osgGA::GUIEventAdapter::KEYUP:
                return true;
*/
            default:
                return false;
        }
    }

    virtual void accept(osgGA::GUIEventHandlerVisitor& v) {
            v.visit(*this);
    }

};


void createKeyboardHandler(osg::Group* tileableNode, osg::Group* tileableParent) {
    if (batch || (viewer==0)) return;
    viewer->getEventHandlerList().push_front(new KeyboardEventHandler(tileableNode, tileableParent));
}

// start without a generator and allocate it on demand
// => avoid creating a RNG with time seed and re-seeding
static TRandomMersenne* rng = 0;

static inline TRandomMersenne* getRNG() {
    // use most frequent condition in if test for CPU branch prediction
    if (rng!=0) return rng;
    randomSeed = time(0);
    rng = new TRandomMersenne(randomSeed);
    ostringstream os;
    os << "Random seed=" << randomSeed;
    Utility::setStatus(os.str(), BOTTOM_LEFT, true);
    return rng;
}

long RandomInit() {
    randomSeed = time(0);
    if (rng==0) rng = new TRandomMersenne(randomSeed);
    else rng->RandomInit(randomSeed);
    ostringstream os;
    os << "Random seed=" << randomSeed;
    Utility::setStatus(os.str(), BOTTOM_LEFT, true);
    return randomSeed;
}

void RandomInit(long seed) {
    randomSeed = seed;
    if (rng==0) rng = new TRandomMersenne(randomSeed);
    else rng->RandomInit(randomSeed);
    ostringstream os;
    os << "Random seed=" << randomSeed;
    Utility::setStatus(os.str(), BOTTOM_LEFT, true);
}

int IRandom(int min, int max) {
    return getRNG()->IRandom(min, max);
}

double Random() {
    return getRNG()->Random();
}

long BRandom() {
    return getRNG()->BRandom();
}

double NRandom() {
    double x;
    do {
        x = Random();
    } while (x==0.0);
    x = sqrt(-2.0 * log(x));
    if (Random() > 0.5) return x * cos((2.0*3.14159265358979323846) * Random());
    return x * sin((2.0*3.14159265358979323846)* Random());
}


struct FileLogger : public Logger {

    ofstream *file;
    
    FileLogger(const std::string& id, const std::vector< std::string > & names ) : Logger(id, names) {

        // First, build a file name. Use the random seed for uniqueness.
        ostringstream os;
        os << id << "_" << randomSeed << ".txt";
        
        file = new ofstream(os.str().c_str());
        (*file) << "# Crogai - genebasic v0.2 experiment result file." << endl;
        (*file) << "# The following parameters were used:" << endl;
        (*file) << "#   Random seed: " << randomSeed << endl;
        (*file) << "#   Agent initial mass: " << AgentCore::getDefaultInitialMass() << endl;
        (*file) << "#   Energy footprint ratio: " << AgentCore::getDefaultEnergyFootprintRatio() << endl;
        (*file) << "#   Agent maximum energy: " << AgentCore::getDefaultMaxEnergy() << endl;
        (*file) << "#   Agent self-maintainance energy cost per time unit: " << AgentCore::getDefaultSelfMaintainEnergyCost() << endl;
        (*file) << "#   Agent maximum force: " << AgentCore::getDefaultMaxForce() << endl;
        (*file) << "#   Agent maximum speed: " << AgentCore::getDefaultMaxSpeed() << endl;
        (*file) << "#   Agent integration time step: " << agentsIntegrationTimeStep << endl;
        (*file) << "#   Agent AI time step: " << agentsAITimeStep << endl;
        (*file) << "#   Spore cloud diffusion integration time step: " << cloudIntegrationTimeStep << endl;
        (*file) << "#   Spore cloud diffusion coefficient: " << cloud->getDiffusionCoefficient() << endl;
        (*file) << "#   World gravity: " << gravity << endl;
        (*file) << "# All other parameters have their default values in v0.2 of the project" << endl;
        (*file) << endl;
        for (std::vector<std::string>::const_iterator it = names.begin(); it!=names.end(); ++it) {
            if (it!=names.begin()) (*file) << " ";
            (*file) << *it;
        }
        (*file) << endl;
    }

    ~FileLogger() {
        file->close();
        delete file;
    }
    
    virtual void record() {
        for (std::vector<double>::const_iterator it = this->begin(); it!=this->end(); ++it) {
            if (it!=this->begin()) (*file) << " ";
            (*file) << *it;
        }
        (*file) << endl;
    }
    
};


Logger * createLogger( const std::string& id, const std::vector< std::string > & names )
{
    return new FileLogger(id, names);
}


struct BatchThread : public OpenThreads::Thread {
    
    virtual void run() {
        while (true) {
            // Directly copied from man page!
            fd_set rfds;
            struct timeval tv;
            int retval;
    
            /* Watch stdin (fd 0) to see when it has input. */
            FD_ZERO(&rfds);
            FD_SET(0, &rfds);
            // Poll 5 times a sec, ridiculous for a computer, too fast for human
            tv.tv_sec = 0;
            tv.tv_usec = 200000;
    
            retval = select(1, &rfds, NULL, NULL, &tv);

            // break when some character is present
            if (retval>0) break;
            
        }
        // stop simu loop
        if (simu) simu->stop();
    }

};



void mainLoop() {
    if (batch) {
        OpenThreads::Thread::Init(); // allow threading if it wasn't done before
        BatchThread bthread;
        bthread.start(); // loop for input
        if (simu) simu->start(false); // use main thread. Till stopped, either by batch thread or by simu event
        bthread.cancel(); // whatever the result, stop batch now
    }
    else {
        if (simu) simu->start(); // in its own thread
        // use this thread for OSG main loop
        while( !viewer->done() )
        {
            viewer->sync();
            viewer->update();
            viewer->frame();
        }
        // wait for all OSG threads to sync
        viewer->sync();
    }
        
    // flush all pending simu events (cyclic...)
    if (simu) {
        simu->flush();
        // wait for simu to stop (if not stopped already)
        simu->stop(true);
    }
}



} // end namespace

