/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osg/Geometry>

#include <osgDB/ReadFile>
#include <osg/Material>
#include <osg/Geode>

#include <iostream>
#include <fstream>
#include <sstream>

#include <time.h>
#include <math.h>
#include <unistd.h>

#include <map>

#include "Terrain.h"
#include "Environment.h"
#include "KeepUpAttitudeCorrection.h"
#include "Obstacle.h"
#include "TwoLayerPerceptron.h"
#include "NeuronalAgent.h"
#include "DrawableSporeCloud.h"
#include "Utility.h"
#include "ExplicitAIAgent.h"


/*


TODO

Introduce energy sources, perhaps the bushes => provide an input to this open
dissipative system, which is at present introduced by reproduction
cost < initial energy.

*/







using namespace std;
using namespace Environment;

Terrain* theFloor = 0;
double istep = 0.04; // 25 f/s
double aistep = 0.15;
double cloudstep = 0.3;

static const double initialEnergy = 500.0;
static const double reproductionEnergy = 1000.0;
static const int maxReproductionNeighbors = 5;
static const int maxPopulationSize = 300;
static const double maxForce = 20;
static const double worldGravity = 17;
static const double maxSpeed = 10;
static const double energyFootprintRatio = 0.001;

osg::Group* root = 0, *metaroot = 0;

/// Adds a new agent to the scene graph.
/// Also used as callback to be executed at a safe place for OSG.
static void addNewAgent(CrogaiAgent* agent) {

    agent->addAvatarToGroup(root);

    // add roughly one fourth of the agents to the locality database
    // the others keep their own neighborhood
    //if (Utility::Random() < 0.25)
    agent->addToLocalityDatabase();

    Agent::allAgentsAccessLock();
    Agent::allAgents.push_back(agent);
    Agent::allAgentsAccessUnlock();
    
    agent->aiEvent = simu->postCyclicAt<Agent>(
        // add jitter so all ai computations not done at the same time
        simu->getTime() + aistep * Utility::Random(),
        aistep, agent, &Agent::aiStep);
    
    agent->integrationEvent = simu->postCyclicAt<Agent>(
        // add jitter so all integration step computations not done at the same time
        simu->getTime() + istep * Utility::Random(),
        istep, agent, &Agent::integrationStep);
    
}

/*
struct ExplicitAIAgent : public CrogaiAgent {

    static DrawableSporeCloud::ColoredGenome *genome;
    
    ExplicitAIAgent(int specie, bool addToGlobalAgentList = true) : CrogaiAgent(specie, addToGlobalAgentList) {

        if (genome==0) {
            genome = new DrawableSporeCloud::ColoredGenome;
            genome->setHalfLife(60.0);
            genome->r = 1.0f;
            genome->g = 0.0f;
            genome->b = 0.0f;
        }
        
    }
    

    void updateAI( const double elapsedTime, const double currentTime) {
        CrogaiAgent::updateAI(elapsedTime,currentTime);
        
        osg::Vec3d localCenter(0,0,0);
        osg::Vec3d forwardForce(0,0,0);
        osg::Vec3d obsForce(0,0,0); // obstacle avoidance
        osg::Vec3d seekFleeForce(0,0,0); // obstacle avoidance
        
        // avoidance force. Avoid only the closest object
        // Since neighbors are returned sorted, this is easy
        osg::Vec3d avoidForce(0,0,0);
        
        int nSeekFlee = 0; // number of neighbor preys + predators
        int n = 0; // number of neighbor agents
        
        findNeighbors(searchRadius);
        neighborsAccessLock();
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

                if (wrappedDiffVector(it->second->getPosition(), getPosition()) * forward < 0)
                    continue; // see only what's in front of us

                Agent* neighbor = dynamic_cast<Agent*>(it->second);

                // if not an agent, maybe it is an obstacle?
                if (!neighbor) {
                    Obstacle* obs = dynamic_cast<Obstacle*>(it->second);
                    // Avoid obstacles, but only the closest one for which there is a collision risk
                    if (obs!=0) {
                        avoidForce = steerToAvoid(obs->getNode(), searchRadius);
                        break;
                    }
                    continue;
                }


                ExplicitAIAgent* sa = dynamic_cast<ExplicitAIAgent*>(neighbor);
                // Escape from unknown things from outer space!!!
                if (sa==0) {
                    steeringForce += steerForEvasion(neighbor);
                }
                // One of our mates. Boidizing...
                else if (sa->specie == specie) {
                    // Check for risk of collision
                    avoidForce = steerToAvoid(neighbor, searchRadius);
                    
                    // Collision risk => no need to look for further AI
                    if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0))
                        break;
    
                    steeringForce += steerForFlee(neighbor->getPosition()) / (it->first+0.001);
                    forwardForce += neighbor->forward;
                    localCenter += neighbor->getPosition();
                    ++n;
                }
                // Hey, another guy! Let's try to catch it unaware...
                else {
                    // hehe, we are behind, it can't see us
                    if (sa->forward * forward < 0) {
                        seekFleeForce += steerForPursuit(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                    // Oops, avoid fight
                    else {
                        seekFleeForce += steerForEvasion(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                }
        }
        neighborsAccessUnlock();

        // Collision risk? Priority over anything else
        if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0)) {

            // Don't try to avoid by digging a hole, it doesn't work!
            // When not flying, project avoidance direction on the horizontal plane
            if (altitude<radius) {
                steeringForce = avoidForce - floorNormal * (avoidForce*floorNormal);
                steeringForce *= sqrt(getMaxForceSquared() / steeringForce.length2());
            }
            else steeringForce = avoidForce;

            return;
        }

        // don't move to the origin when no other agent present
        // In this case, move at random, but only if not already doing something else
        if ((n==0) && (nSeekFlee==0)) steeringForce += steerForWander(0.4);

        // Otherwise, agent behavior when in company of friends
        // - steer to avoid neighbors
        // - steer toward the local center
        // - and also to keep the local group direction
        else if (n>0) {
            steeringForce /= (double)n; // average of avoidance vectors
            steeringForce += steerForSeek(localCenter/(double)n) * 2; // local center
            steeringForce += forwardForce /(double)n  -  forward; // local group direction
        }

        // prey/predator avoidance?
        if (nSeekFlee>0) {
            steeringForce += seekFleeForce * 10; // give it more weight than boidizing
        }
        
        // If flying, try to avoid floor, up to max height were we're "safe"
        if (altitude>radius) {
            if ( altitude < radius*5.0) 
                steeringForce += floorNormal * getMaxForce();
        }
        
        // Otherwise, have a 1% chance to take off at random, supposing gravity doesn't bring us back
        else {
            if (1.0 * Utility::Random() < 0.01) steeringForce += floorNormal * getMaxForce();
        }
        

    }

    virtual void winnerAction(CrogaiAgent* looser) {
        CrogaiAgent::winnerAction(looser);
        
        //cout << "Team " << specie << " score is now " << scores[specie] << endl;

        if (energy<reproductionEnergy) return; // not enough energy
        int popSize;
        allAgentsAccessLock();
        popSize = allAgents.size();
        allAgentsAccessUnlock();
        if (popSize>maxPopulationSize) return;
        
        pouf();
        
        // Haha! Enough energy to spawn!

        // Don't add child to list now in graphics mode
        CrogaiAgent* agent = new ExplicitAIAgent(specie, false);
        updateChild(agent); // parent positioning code
        // But delay this to next OSG update
        if (batch) addNewAgent(agent);
        else viewer->post(&addNewAgent, agent);
        
        ostringstream os;
        os << "Scores: ";
        for (map<int, int>::iterator it = scores.begin(); it != scores.end(); ++it) {
            if (it!=scores.begin()) os << ", ";
            os << it->first << ":" << it->second;
        }
        os << ", energy=" << energy << ". AI";
        Utility::setStatus(os.str(), Utility::TOP_LEFT);
        
        //agent->body->setColor(osg::Vec4(1,0,1,1));
    }
    
    virtual void looserAction(CrogaiAgent* winner) {
        CrogaiAgent::looserAction(winner);
        
        int popSize;
        allAgentsAccessLock();
        popSize = allAgents.size();
        allAgentsAccessUnlock();

        ostringstream os;
        os << "Pop=" << popSize;
        Utility::setStatus(os.str(), Utility::TOP_RIGHT);
    }
    
    void pouf() {
        cloud->addSpore(getPosition().x(), getPosition().y(), getPosition().z(), genome, 1.0);
    }
};
DrawableSporeCloud::ColoredGenome *ExplicitAIAgent::genome = 0;
*/

class SpawningNeuronalAgent : public NeuronalAgent {
public:

    int generation;
    DrawableSporeCloud::ColoredGenome *genome;

    // parent constructor
    SpawningNeuronalAgent(int nhidden, int specie, AggregateStrategy neighborsStrategy = Average, AggregateStrategy forcesStrategy = Average, bool addToGlobalAgentList = false)
    : NeuronalAgent(nhidden, specie, neighborsStrategy, forcesStrategy, addToGlobalAgentList) {
        generation = 0;
        genome = new DrawableSporeCloud::ColoredGenome;
        genome->setHalfLife(60.0);
    }

    // Overload to return an object of this type
    CrogaiAgent * spawn(bool addToGlobalAgentList = true)
    {
        SpawningNeuronalAgent* ret = new SpawningNeuronalAgent(neighborTLP.getNHidden(), specie, aggregateAgents, aggregateForces, addToGlobalAgentList);
    
        // introduce mutations from parent
        updateChild(ret);

        ret->generation = generation + 1;
        ret->genome->r = 0.0f;
        ret->genome->g = sin(generation*6.2831853*0.1);
        ret->genome->b = cos(generation*6.2831853*0.1);
        
        // child is ready
        return ret;
    }
    
    
    virtual void winnerAction(CrogaiAgent* looser) {
        CrogaiAgent::winnerAction(looser);
        
        //cout << "Team " << specie << " score is now " << scores[specie] << endl;

        if (energy<reproductionEnergy) return; // not enough energy
        int popSize;
        allAgentsAccessLock();
        popSize = allAgents.size();
        allAgentsAccessUnlock();
        if (popSize>maxPopulationSize) return;

        pouf();
        
        // Haha! Enough energy to spawn!

        // Don't add child to list now in graphics mode
        CrogaiAgent* agent = spawn(batch);
        // But delay this to next OSG update
        if (batch) addNewAgent(agent);
        else viewer->post(&addNewAgent, agent);
        
        //agent->body->setColor(osg::Vec4(1,0,1,1));
    }
    
    static osg::Box* commonShape;
    osg::Node *buildAvatar()
    {
        // Share the body of these agents across all members of the class, thanks to ref counting
        if (!commonShape) commonShape = new osg::Box(osg::Vec3(0,0,0),1,2,0.5);
        
        body = new osg::ShapeDrawable(commonShape);
    
        // Allow the color to be changed at run time
        body->setUseDisplayList( false );
        osg::Vec4 color(0, sin(generation*6.2831853*0.1), cos(generation*6.2831853*0.1), 1.0);
        body->setColor(color);

/*        osg::StateSet* stateset = body->getOrCreateStateSet();
        stateset->setMode(GL_BLEND,osg::StateAttribute::ON);
        //stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
*/
            
        // OSG used geodes as generic containers. Bridge from osg::Shape to osg::Node
        osg::Geode* geode = new osg::Geode();
        geode->addDrawable(body);
        return geode;
    }


    void pouf() {
        cloud->addSpore(getPosition().x(), getPosition().y(), getPosition().z(), genome, 1.0);
    }
};

osg::Box* SpawningNeuronalAgent::commonShape = 0;

/*double tanhDerivativeAsF(double t) {
    return 1.0 - t*t;
}
*/


int main(int argc, char *argv[])
{

    // TODO batch = argv[...]
    if (argc>1) batch = true;

//    osg::Referenced::setDeleteHandler( new ThreadSafeDeleteHandler);
    osg::Referenced::setThreadSafeReferenceCounting(true);
    
    // setup env variables
    agentsIntegrationTimeStep = istep;
    agentsAITimeStep = aistep;
    cloudIntegrationTimeStep = cloudStep;

    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();

    if (!batch) createViewer("Crowd - GA - AI, evolution");

    root = new osg::Group();
    
    //root->addChild(createFloor(200));

//    theFloor = new Terrain(800, 100, 7, 0.3, 3.0);
//    theFloor = new Terrain(400, 50, 7, 0.3, 3.0);
    theFloor = new Terrain(500, 100, 7, 0.35, 3.5);
    root->addChild(theFloor);
    TerrainAgent::terrain = theFloor;

        
    if (!batch) {
        metaroot = new osg::Group();
        metaroot->addChild(root);
        osg::Group * wholeScene = new osg::Group();
        wholeScene->addChild(metaroot);
        Utility::createKeyboardHandler(root, metaroot);
        
        viewer->setSceneData( wholeScene );
        viewer->realize();
    }

    cloud = new DrawableSporeCloud(theFloor, 32, 32, 10, 75.0, 0.001);
    cloud->setDefaultAttributes("smoke.rgb", false, false);
    
    simu->postCyclicAt<SporeCloud>(cloudIntegrationStep * Utility::Random(), cloudIntegrationStep, cloud, &SporeCloud::step);

    osg::Geode* geode = new osg::Geode();
    geode->addDrawable(cloud);
    root->addChild(geode);

        
    //Localized::setDefaultLocalityMode(Localized::Mixed);
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        100,100,50);

    if (batch) simu->setTimeFactor(100000.0);

// Obstacle code
    const int nobstacles = 75;
    for (int i=0; i<nobstacles; ++i) {
        double x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        
        // Position Obstacle at floor surface
        Obstacle *obs = new Obstacle(x, y, theFloor->getFloorZ(x,y));
        obs->addAvatarToGroup(root);

        // Obstacles update their neighbor lists to help agents, in mixed mode only
        /*simu->postCyclicAt(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),
            aistep, obs, &Obstacle::updateNeighbors, 30.0);
        */
    }

    Environment::gravity = worldGravity;
    AgentCore::setDefaultMaxForce(maxForce);
    AgentCore::setDefaultMaxSpeed(maxSpeed);
    AgentCore::setDefaultEnergyFootprintRatio(energyFootprintRatio);
    AgentCore::setDefaultInitialEnergy(initialEnergy);

// Create a neuronal agent, with trainer weights, to serve as ancestor for all others
    SpawningNeuronalAgent *na = new SpawningNeuronalAgent(50, 1);//, NeuronalAgent::WinnerTakeAll, NeuronalAgent::Average);
    na->monitorError = false;
/*    na->aloneTLP.transfer = &tanh;
    na->aloneTLP.transferDerivativeAsF = &tanhDerivativeAsF;
    na->neighborTLP.transfer = &tanh;
    na->neighborTLP.transferDerivativeAsF = &tanhDerivativeAsF;
*/

    ifstream tlpfile;
    tlpfile.open("tlp.txt");
    if (tlpfile.good()) {
        tlpfile >> na->aloneTLP;
        tlpfile >> na->neighborTLP;
    }
    tlpfile.close();

    // set mutation rate to create children
    na->setSpawnParameters(0.1, 0.1);
    
// Agent code
    const int nagents = 200;
    for (int i=0; i<nagents; ++i) {
        CrogaiAgent *agent;
        if (i<nagents/2) agent = na->spawn(); // create mutated offspring, add it to agent list
        else agent = new ExplicitAIAgent(2);  // create fixed AI
        double x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();

        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));

        // Add avatar even in batch mode => neighbors detection needs it
        // TODO: remove this dep
        agent->addAvatarToGroup(root);
        
        agent->aiEvent = simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),
            aistep, agent, &Agent::aiStep);
        
        agent->integrationEvent = simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            istep * Utility::Random(),
            istep, agent, &Agent::integrationStep);

        // add roughly one fourth of the agents to the locality database
        // the others keep their own neighborhood
        //if (Utility::Random() < 0.25)
        agent->addToLocalityDatabase();
        
    }

/*        
    // Provide an initial list of neighbors to all agents
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }
    // Second order search
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }
*/

    simu->start(); // in its own thread

    if (batch) {
        // read on input
        char c = 0;
        while (read(0, &c, sizeof(char))) {
            switch (c) {
                case 27: exit(0);
                default:
                    cout << "Unknown batch command: " << c << endl;
                    break;
            }
        }
        exit(1); // error in reading?
    }
    
    while( !viewer->done() )
    {
        viewer->sync();
        viewer->update();
        viewer->frame();
    }
    // wait for all OSG threads to sync
    viewer->sync();

    // flush all pending simu events (cyclic...)
    simu->flush();
    
    // wait for simu to stop
    simu->stop(true);

    exit(0); // bug in osg/openthreads at object deletion. Use exit instead of return
    return 0;
}



