/*
// ----------------------------------------------------------------------------
//
//
// OpenSteer -- Steering Behaviors for Autonomous Characters
//
// Copyright (c) 2002-2003, Sony Computer Entertainment America
// Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// ----------------------------------------------------------------------------

// 30-11-04: by Nicolas Brodu: Made the bins cyclic in X and Y.
//           This is adapted for proximity computation on a surface "tore", 
//           where the height in Z is considered infinite.
// 27-01-05: by Nicolas Brodu: converted the whole thing to use double
//           instead of float. Rational is double are better supported
//           on current CPU.

*/

/* ------------------------------------------------------------------ */
/*                                                                    */
/*                   Locality Query (LQ) Facility                     */
/*                        Cyclic in X and Y                           */
/*                                                                    */
/* ------------------------------------------------------------------ */
/*

    NB043011:
    
    These routines are compatible with the LQ facility and share the
    same API.
    Just call the XYCyclic versions below instead of the standard versions
    to use the cyclic feature. Arguments are the same.

    Please refer to the lq.h header file for comments.
    
*/

#ifndef _lqXYCyclic_h
#define _lqXYCyclic_h

#include "lq.h"

#ifdef __cplusplus
extern "C"
{
#endif

void lqXYCyclicUpdateForNewLocation (lqDB* lq,
                                     lqClientProxy* object,
                                     double x, double y, double z);

void lqXYCyclicMapOverAllObjectsInLocality (lqDB* lq,
                                            double x, double y, double z,
                                            double radius,
                                            lqCallBackFunction func,
                                            void* clientQueryState);

void* lqXYCyclicFindNearestNeighborWithinRadius (lqDB* lq,
                                                 double x, double y, double z,
                                                 double radius,
                                                 void* ignoreObject);

lqBin* lqXYCyclicBinForLocation (lqDB* lq, double x, double y, double z);


#ifdef __cplusplus
}
#endif

#endif /* _lqXYCyclic_h */
