#!/usr/bin/perl

sub help() {
    print "Missing or bad arguments: argument: statistics_XXX.txt [statistics_YYY.txt ...]\n";
    print "\n";
    print "This script should be run on the statistics log files. It uses the corresponding populations file (same random seed) to check on which species went extinct, and monitor how much generations were produced before extinction.\n";
    exit(1);
};

if ($#ARGV < 0) {
    help();
}

$meanGenerationsTimeRatio = 0;
$varGenerationsTimeRatio = 0;

$totalExtinctionNumber = 0;

# loop over file names
foreach $fileName (@ARGV) {

    if (!($fileName =~ /.*statistics.*/)) {
        print "Incorrect file name: $fileName\n";
        help();
    }
    
    open(FILE,"<$fileName") or die "Cannot open $fileName for reading.\n";
    
    $popFile = $fileName;
    $popFile =~ s/statistics/populations/g;
    open(POPFILE,"<$popFile") or die "Cannot open $popFile for reading.\n";

$first = 0;
@popEntry = ();
@popTimes = ();
POPLOOP: while (<POPFILE>) {

    chomp;

    /^#.*/ && next POPLOOP;
    /^$/ && next POPLOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        @names = split;

        # force allocation of 0 max time for all species
        for ($i=0; $i<$#names; ++$i) {$popTimes[$i] = 0;}
        next POPLOOP;
    }
    
    #discard time => check only on extinctions. Use it here
    @popEntry = split;
    $time = shift @popEntry;
    
    for ($i=0; $i<=$#popEntry; ++$i) {
        if ($popEntry[$i] != 0) {$popTimes[$i] = $time;}
    }
}
close(POPFILE);

# force allocation of 0 for all species
@generations = ();
for ($i=0; $i<=$#popEntry; ++$i) {$generations[$i] = 0;}

$first = 0;
LOOP: while (<FILE>) {

    chomp;

    /^#.*/ && next LOOP;
    /^$/ && next LOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        next LOOP;
    }

    @entry = split;
    
    $gen = shift @entry;
    
    for ($i=0; $i<=$#generations; ++$i) {
        $j = $i * 8;
        if ($entry[$j] != 0) {$generations[$i] = $gen;}
    }
}
close(FILE);

print "# generations number/time (e=extinct, a=alive):";
for ($i=0; $i<=$#generations; ++$i) {
    # only monitor extinct species, final explosion for surviving specie has too much influence
    # let's compare what's comparable!
    if ($popEntry[$i]==0) {
        print " $generations[$i]e/$popTimes[$i]";
        $ratio = $generations[$i] / $popTimes[$i];
        $meanGenerationsTimeRatio += $ratio;
        $varGenerationsTimeRatio += $ratio * $ratio;
        $totalExtinctionNumber ++;
    } else {
        print " $generations[$i]a/$popTimes[$i]";
    }
}
print "\n";

} # end of file loop

# finish computations
$meanGenerationsTimeRatio /= $totalExtinctionNumber;
$varGenerationsTimeRatio = sqrt( ($varGenerationsTimeRatio / $totalExtinctionNumber) - $meanGenerationsTimeRatio * $meanGenerationsTimeRatio );

print "Mean generations number/time ratio (*100) before extinction: ".($meanGenerationsTimeRatio*100).", std. dev: ".($varGenerationsTimeRatio*100)."\n";
