#!/usr/bin/perl

if ($#ARGV < 0) {
    print "Missing or bad arguments: argument: transientTimes1.txt [transientTimes2.txt ...]\n";
    print "\n";
    print "This script should be run on the transient times log files. It computes statistics over all the files in argument.\n";
    exit(1);
}

$meanExtinctionNumber = 0;
$varExtinctionNumber = 0;
$meanExtinctionTime = 0;
$varExtinctionTime = 0;
$meanExplosionTime = 0;
$varExplosionTime = 0;

$totalExtinctionNumber = 0;
$totalExplosionNumber = 0;

# number of extinctions:
# time before extinction:
# time before explosion:

# loop over file names
foreach $fileName (@ARGV) {

    open(FILE,"<$fileName") or die "Cannot open $fileName for reading.\n";

$first = 0;

$extinctionNumber = 0;

LOOP: while (<FILE>) {

    chomp;

    /^#.*/ && next LOOP;
    /^$/ && next LOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        next LOOP;
    }

    print "# $_\n";
    @times = split;

    $inExctinctions = 1;
    TIMELOOP: for ($i=0; $i<=$#times; ++$i) {
        if ($times[$i]==0) {
            $inExctinctions = 0;
        }
        if ($inExctinctions == 1) {
            $extinctionNumber++;
            $totalExtinctionNumber++;
            
            $meanExtinctionTime += $times[$i];
            $varExtinctionTime += $times[$i] * $times[$i];
            next TIMELOOP;
        }
        if ($times[$i]!=0) {
            $meanExplosionTime += $times[$i];
            $varExplosionTime += $times[$i] * $times[$i];
            $totalExplosionNumber++;
        }
    }
    

}
close(FILE);

$meanExtinctionNumber += $extinctionNumber;
$varExtinctionNumber += $extinctionNumber * $extinctionNumber;

} # end of file loop

# finish computations
$numFiles = 1 + $#ARGV;

$meanExtinctionNumber /= $numFiles;
$varExtinctionNumber = sqrt( ($varExtinctionNumber / $numFiles) - $meanExtinctionNumber * $meanExtinctionNumber );

$meanExtinctionTime /= $totalExtinctionNumber;
$varExtinctionTime = sqrt( ($varExtinctionTime / $totalExtinctionNumber) - $meanExtinctionTime * $meanExtinctionTime );

$meanExplosionTime /= $totalExplosionNumber;
$varExplosionTime = sqrt( ($varExplosionTime / $totalExplosionNumber) - $meanExplosionTime * $meanExplosionTime );

print "Mean number of extinctions: $meanExtinctionNumber, standard deviation: $varExtinctionNumber\n";
print "Mean time before extinction: $meanExtinctionTime, standard deviation: $varExtinctionTime\n";
print "Mean time before explosion: $meanExplosionTime, standard deviation: $varExplosionTime\n";
